// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hints

import (
	"bytes"
	"fmt"
	"net/http"
	"net/http/httptest"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

var _ = Describe("Hint Request validation tests", func() {
	timeMillis1 := int64(12345)
	timeMillis2 := int64(67890)

	// ----- Unmarshal testing --------------------

	It("Ensure json can be properly unmarshaled into the structure", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, 
			"device_type":"WINDOWS", "deployment_type":"ENDPOINT",
			"version":1, "last_update_time":12345,
            "last_live_query_time":67890, "remote_rules_required":false,
            "audit_hints":["CONFER_UPLOAD"], "audit_last_update_time": 12345, "audit_last_live_query_time": 67890}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).To(BeNil())

		Expect(jhr.OrgKey).To(BeEquivalentTo("N86QADZN"))
		Expect(jhr.DeviceGroupId).To(BeEquivalentTo(123))
		Expect(jhr.DeviceId).To(BeEquivalentTo(456))
		Expect(jhr.DeviceType).To(BeEquivalentTo("WINDOWS"))
		Expect(jhr.DeploymentType).To(BeEquivalentTo("ENDPOINT"))
		Expect(*jhr.Version).To(BeEquivalentTo(1))
		Expect(*jhr.LastUpdateTime).To(BeEquivalentTo(timeMillis1))
		Expect(*jhr.LastLiveQueryTime).To(BeEquivalentTo(timeMillis2))
		Expect(jhr.RemoteRulesRequired).To(BeEquivalentTo(false))
		check := []string{"CONFER_UPLOAD"}
		Expect(jhr.AuditHints).To(BeEquivalentTo(check))
		Expect(*jhr.AuditLastUpdateTimestamp).To(BeEquivalentTo(timeMillis1))
		Expect(*jhr.AuditLastLiveQueryTimestamp).To(BeEquivalentTo(timeMillis2))
	})

	It("Ensure bare-minimum json can be properly unmarshaled into the structure", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).To(BeNil())

		Expect(jhr.OrgKey).To(BeEquivalentTo("N86QADZN"))
		Expect(jhr.DeviceGroupId).To(BeEquivalentTo(123))
		Expect(jhr.DeviceId).To(BeEquivalentTo(456))
		Expect(jhr.DeviceType).To(BeEquivalentTo(""))
		Expect(jhr.DeploymentType).To(BeEquivalentTo(""))
		Expect(jhr.Version).To(BeNil())
		Expect(jhr.LastUpdateTime).To(BeNil())
		Expect(jhr.LastLiveQueryTime).To(BeNil())
		Expect(jhr.RemoteRulesRequired).To(BeEquivalentTo(false))
		Expect(jhr.AuditHints).To(BeNil())
	})

	// NOTE: If 'device_type' or 'deployment_type' is not specified, we expect an empty string value

	// ----- Body Structure testing --------------------
	// NOTE: These deal with body structure validation, not content validation

	It("Ensure we reject a missing org_key", func() {
		rawJSON := `{"device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'JsonHintRequest.OrgKey' Error:Field validation for 'OrgKey' failed on the 'required' tag"))
	})

	// NOTE: Rejecting an invalid org_key handled in hint_contoller tests

	It("Ensure we reject a missing device_group_id", func() {
		// NOTE: if not provided, defaults to 0; we trap this
		rawJSON := `{"org_key":"N86QADZN", "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'JsonHintRequest.DeviceGroupId' Error:Field validation for 'DeviceGroupId' failed on the 'min' tag"))
	})

	It("Ensure we reject a non-int device_group_id", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":"12J3", "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.device_group_id of type int64"))
	})

	It("Ensure we reject a missing device_id", func() {
		// NOTE: if not provided defaults to 0; we trap this
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123 , "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'JsonHintRequest.DeviceId' Error:Field validation for 'DeviceId' failed on the 'min' tag"))
	})

	It("Ensure we reject a non-int device_id", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id": "456HJ", "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.device_id of type int64"))
	})

	// NOTE: 'device_type' or 'deployment_type' are optional

	// NOTE: 'version' is optional

	It("Ensure we reject a non-int version", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":"Q", "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.version of type int"))
	})

	// NOTE: last_update_time is optional

	It("Ensure we reject a non-int last update time", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":"2023-02-14T12:13:14", "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.last_update_time of type int64"))
	})

	// NOTE: last_live_query_time is optional

	It("Ensure we reject a non-int last livequery time", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":"2023-02-14T12:13:14", "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.last_live_query_time of type int64"))
	})

	// NOTE: audit_hints is optional

	It("Verify a missing audit_hints is not there", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).To(BeNil())
		Expect(jhr.AuditHints).To(BeNil())
	})

	It("Ensure we accept an empty-array audit_hints", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false, "audit_hints":[]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).To(BeNil())
	})

	It("Ensure we reject a non-array audit_hints", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false, "audit_hints":"HASH_FORENSICS_LIST"}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.audit_hints of type []string"))
	})

	It("Ensure we reject audit_hints that is not array of string", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false, "audit_hints":[19]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal number into Go struct field JsonHintRequest.audit_hints of type string"))
	})

	// NOTE: audit_last_update_time is optional

	It("Ensure we reject a non-int audit_last_update_time", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false, "audit_last_update_time":"2023-02-18T12:00:00"}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.audit_last_update_time of type int64"))
	})

	// NOTE: audit_last_live_query_time is optional

	It("Ensure we reject a non-int audit_last_live_query_time", func() {
		rawJSON := `{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "version":1, "last_update_time":0, "last_live_query_time":0, "remote_rules_required":false, "audit_last_live_query_time":"2023-02-18T12:00:00"}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/hint/uri", bytes.NewBufferString(rawJSON))
		jhr := JsonHintRequest{}
		err := context.ShouldBindJSON(&jhr)
		Expect(err).ToNot(BeNil())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("json: cannot unmarshal string into Go struct field JsonHintRequest.audit_last_live_query_time of type int64"))
	})

	// ----- DB Conversion testing --------------------
	// NOTE: These validate Unmarshal checks

	It("Ensure we produce proper unmarshaled data)", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, 
			"device_type":"WINDOWS", "deployment_type":"ENDPOINT",
			"version":1, "last_update_time":12345,
            "last_live_query_time":67890, "remote_rules_required":false,
            "audit_hints":["CONFER_UPLOAD"], "audit_last_update_time": 12345, "audit_last_live_query_time": 67890}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.OrgKey).To(BeEquivalentTo("N86QADZN"))
		Expect(hrm.DeviceGroupId).To(BeEquivalentTo(123))
		Expect(hrm.DeviceId).To(BeEquivalentTo(456))
		Expect(hrm.Version).To(BeEquivalentTo(1))
		lut := time.UnixMilli(12345)
		Expect(hrm.LastUpdateTime).To(BeEquivalentTo(&lut))
		llqt := time.UnixMilli(67890)
		Expect(hrm.LastLiveQueryTime).To(BeEquivalentTo(&llqt))
		Expect(hrm.DeviceType).To(BeEquivalentTo(devicetype.WINDOWS))
		Expect(hrm.DeploymentType).To(BeEquivalentTo(deploymenttype.ENDPOINT))
		Expect(hrm.AuditHintResult).ToNot(BeNil())
		Expect(hrm.AuditHintResult.Hints).To(BeEquivalentTo(hinttype.NewHintSet(hinttype.CONFER_UPLOAD)))
		Expect(hrm.AuditHintResult.LastUpdateTime).To(BeEquivalentTo(&lut))
		Expect(hrm.AuditHintResult.LastLiveQueryTime).To(BeEquivalentTo(&llqt))
	})

	It("Ensure we produce proper defaults with only manditory fields)", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeviceType).To(BeEquivalentTo(devicetype.UNKNOWN))
		Expect(hrm.DeploymentType).To(BeEquivalentTo(deploymenttype.UNKNOWN))
		Expect(hrm.Version).To(BeEquivalentTo(BASEMESSAGE.LIVE_QUERY_VERSION))
		Expect(hrm.LastUpdateTime).To(BeNil())
		Expect(hrm.LastLiveQueryTime).To(BeNil())
		Expect(hrm.AuditHintResult).To(BeNil())
	})

	It("Ensure orgkey is properly sanitized to uppercase)", func() {
		data := []byte(`{"org_key":"n86qadzn", "device_group_id":123, "device_id":456,
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.OrgKey).To(BeEquivalentTo("N86QADZN"))
	})

	It("Ensure device type is properly sanitized to uppercase before checking)", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456,
			"device_type":"Windows", "deployment_type":"ENDPOINT",
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeviceType).To(BeEquivalentTo(devicetype.WINDOWS))
	})

	It("Ensure unknown device type is properly handled", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456,
			"device_type":"AMIGA", "deployment_type":"ENDPOINT",
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeviceType).To(BeEquivalentTo(devicetype.DeviceType("AMIGA")))
	})

	It("Ensure whitespace device type is properly converted to UNKNOWN", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456,
			"device_type":"     ", "deployment_type":"ENDPOINT",
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeviceType).To(BeEquivalentTo(devicetype.UNKNOWN))
	})

	It("Ensure deployment type is properly sanitized to uppercase before checking", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456,
			"device_type":"WINDOWS", "deployment_type":"EndPoint",
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeploymentType).To(BeEquivalentTo(deploymenttype.ENDPOINT))
	})

	It("Ensure undefined deployment type is properly converted)", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456,
			"device_type":"WINDOWS", "deployment_type":"CLOUD",
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeploymentType).To(BeEquivalentTo(deploymenttype.DeploymentType("CLOUD")))
	})

	It("Ensure all whitespace deployment type is properly converted to UNKNOWN)", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456,
			"device_type":"WINDOWS", "deployment_type":"    ",
			"last_update_time":12345, "last_live_query_time":67890, "remote_rules_required":false}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.DeploymentType).To(BeEquivalentTo(deploymenttype.UNKNOWN))
	})

	It("verify that audit hint times only exist if we supply a hint set", func() {
		data := []byte(`{"org_key":"N86QADZN", "device_group_id":123, "device_id":456, "last_update_time":0, "last_live_query_time":0, 
			"remote_rules_required":false, "audit_last_live_query_time":123, "audit_last_update_time":456}`)
		hrm := HintRequestMessage{}
		err := hrm.UnmarshalJSON(data)
		Expect(err).To(BeNil())
		Expect(hrm.AuditHintResult).To(BeNil())
	})

	// ----- Utility coverage ----------

	It("Ensure TimePtrToInt64Ptr() returns nil(int64) if nil(time)  is supplied", func() {
		num := int64(12345)
		intTime := time.UnixMilli(num)
		check := TimePtrToInt64Ptr(&intTime)
		Expect(check).To(BeEquivalentTo(&num))

		var num2 *int64
		var t2 *time.Time
		check = TimePtrToInt64Ptr(t2)
		Expect(check).To(BeEquivalentTo(num2))
	})
})
